'''
Pulsar :mod:`~greenio` facilitates the integration of synchronous
third-party libraries into pulsar asynchronous framework.
It requires the :greenlet:`greenlet <>` library.

If you want to understand how integration works but you are unfamiliar with
greenlets, check out the :greenlet:`greenlet documentation <>` first.

This application **does not use monkey patching** and therefore it
works quite differently from implicit asynchronous libraries such as
gevent_. All it does, it provides the user with a set
of utilities for **explicitly** transferring execution from one greenlet
to a another which execute the blocking call in a greenlet-friendly way.

The caller has the responsibility that the blocking call is greenlet-friendly,
i.e. it transfers the control of execution back to the parent greenlet when
needed.

Usage
=======

Lets assume you are building an application which uses pulsar asynchronous
engine and would like to

* either use an external library written in blocking style,
  i.e. without yielding control to the event loop when IO calls are performed.
* or write your client code without dealing with :class:`~asyncio.Future` or
  coroutines, in other words in an implicit asynchronous style. In this way
  your client code can be used on other frameworks just as well.

In both cases, the :class:`~pulsar.apps.greenio` application is what you need.

.. _green-wsgi:

Green WSGI
---------------

Assume you are using pulsar web server and would like to write your application
in an implicit asynchronous mode, i.e. without dealing with futures nor
coroutines, then you can wrap your WSGI ``app`` with the :class:`.GreenWSGI`
utility::

    from pulsar.apps import wsgi, greenio

    green_pool = greenio.GreenPool()
    callable = wsgi.WsgiHandler([wsgi.wait_for_body_middleware,
                                 greenio.GreenWSGI(app, green_pool)],
                                async=True)

    wsgi.WsgiServer(callable=callable).start()

The :class:`.GreenPool` manages a pool of greenlets which execute your
application. In this way, within your ``app`` you can invoke the
:func:`.wait` function when needing to wait for asynchronous results to be
ready.

.. _green-http:

Green Http
-----------------

The :class:`.HttpClient` can be used with greenlets::

    >>> from pulsar.apps.http import HttpClient
    >>> http = HttpClient(green=True)
    >>> http.green
    True

And now you can write synchronous looking code and run it in a separate
greenlet via the :func:`.run_in_greenlet` decorator::

    @greenio.run_in_greenlet
    def example():
        response = http.get('http://bbc.co.uk')
        ...
        return 'done'


and somewhere, in your asynchronous code::

        result = yield example()
        result == 'done'


the :func:`.run_in_greenlet` decorator, execute the function on a child
greenlet without blocking the asynchronous engine. Once the ``example``
function returns, the asynchronous code continue from the ``yield``
statement as usual.


API
======

Wait
----------

.. autofunction:: wait


Run in greenlet
-------------------

.. autofunction:: run_in_greenlet


Green Pool
----------------

.. autoclass:: GreenPool
   :members:
   :member-order: bysource

Green Lock
----------------

.. autoclass:: GreenLock
   :members:
   :member-order: bysource


Green WSGI
----------------

.. autoclass:: GreenWSGI
   :members:
   :member-order: bysource


.. _gevent: http://www.gevent.org/
'''
import threading
import asyncio
import logging
from collections import deque
from functools import wraps

from greenlet import greenlet, getcurrent

from pulsar import isfuture, async
from pulsar import Future, get_event_loop, AsyncObject, is_async


_DEFAULT_WORKERS = 100
_MAX_WORKERS = 1000


class _DONE:
    pass


class GreenletWorker(greenlet):
    pass


def wait(value):
    '''Wait for a possible asynchronous value to complete.
    '''
    current = getcurrent()
    parent = current.parent
    return parent.switch(value) if parent else value


def run_in_greenlet(callable):
    '''Decorator to run a ``callable`` on a new greenlet.

    A ``callable`` decorated with this decorator returns a coroutine
    '''
    @wraps(callable)
    def _(*args, **kwargs):
        green = GreenletWorker(callable)
        # switch to the new greenlet
        result = green.switch(*args, **kwargs)
        # back to the parent
        while is_async(result):
            # keep on switching back to the greenlet if we get a Future
            try:
                result = green.switch((yield from result))
            except Exception as exc:
                result = green.throw(exc)

        return green.switch(result)

    return _


class GreenPool(AsyncObject):
    '''A pool of running greenlets.

    This pool maintains a group of greenlets to perform asynchronous
    tasks via the :meth:`submit` method.
    '''
    worker_name = 'exec'

    def __init__(self, max_workers=None, loop=None):
        self._loop = loop or get_event_loop()
        self._max_workers = min(max_workers or _DEFAULT_WORKERS, _MAX_WORKERS)
        self._greenlets = set()
        self._available = set()
        self._queue = deque()
        self._shutdown = False
        self._waiter = None
        self._logger = logging.getLogger('pulsar.greenpool')
        self._shutdown_lock = threading.Lock()

    @property
    def max_workers(self):
        return self._max_workers

    @max_workers.setter
    def max_workers(self, value):
        value = int(value)
        assert value > 0
        self._max_workers = value

    def submit(self, func, *args, **kwargs):
        '''Equivalent to ``func(*args, **kwargs)``.

        This method create a new task for function ``func`` and adds it to
        the queue.
        Return a :class:`~asyncio.Future` called back once the task
        has finished.
        '''
        with self._shutdown_lock:
            if self._shutdown:
                raise RuntimeError(
                    'cannot schedule new futures after shutdown')
            future = Future(loop=self._loop)
            self._put((future, func, args, kwargs))
            return future

    def shutdown(self, wait=True):
        with self._shutdown_lock:
            self._shutdown = True
            self._put(None)
            if wait:
                self._waiter = Future(loop=self._loop)
                return self._waiter

    # INTERNALS
    def _adjust_greenlet_count(self):
        if (not self._shutdown and not self._available and
                len(self._greenlets) < self._max_workers):
            green = GreenletWorker(self._green_run)
            self._greenlets.add(green)
            self.logger.debug('Num greenlets: %d', len(self._greenlets))
            green.switch()
        return self._available

    def _put(self, task):
        # Run in the main greenlet of the evnet-loop thread
        self._queue.appendleft(task)
        self._check_queue()

    def _check_queue(self):
        # Run in the main greenlet of the event-loop thread
        if not self._adjust_greenlet_count():
            self.logger.debug('No greenlet available')
            return self._loop.call_soon(self._check_queue)
        try:
            task = self._queue.pop()
        except IndexError:
            return
        async(self._green_task(self._available.pop(), task), loop=self._loop)

    def _green_task(self, green, task):
        # Coroutine executing the in main greenlet
        # This coroutine is executed for every task put into the queue

        while task is not _DONE:
            # switch to the greenlet to start the task
            task = green.switch(task)

            # if an asynchronous result is returned, yield from
            while is_async(task):
                try:
                    task = yield from task
                except Exception as exc:
                    # This call can return an asynchronous component
                    task = green.throw(exc)

    def _green_run(self):
        # The run method of a worker greenlet
        task = True
        while task:
            green = getcurrent()
            parent = green.parent
            assert parent
            # add greenlet in the available greenlets
            self._available.add(green)
            task = parent.switch(_DONE)  # switch back to the main execution
            if task:
                future, func, args, kwargs = task
                try:
                    result = func(*args, **kwargs)
                except Exception as exc:
                    future.set_exception(exc)
                else:
                    future.set_result(result)
            else:  # Greenlet cleanup
                self._greenlets.remove(green)
                if self._greenlets:
                    self._put(None)
                elif self._waiter:
                    self._waiter.set_result(None)
                    self._waiter = None
                parent.switch(_DONE)


class GreenLock:
    '''A Locking primitive that is owned by a particular greenlet
    when locked.The main greenlet cannot acquire the lock.

    A primitive lock is in one of two states, 'locked' or 'unlocked'.

    It is created in the unlocked state. It has two basic methods,
    :meth:`.acquire` and :meth:`.release. When the state is unlocked,
    :meth:`.acquire` changes the state to locked and returns immediately.

    When the state is locked, :meth:`.acquire` blocks the current greenlet
    until a call to :meth:`.release` changes it to unlocked,
    then the :meth:`.acquire` call resets it to locked and returns.
    '''
    def __init__(self, loop=None):
        self._loop = loop or get_event_loop()
        self._local = threading.local()
        self._local.locked = None
        self._queue = deque()

    def locked(self):
        ''''Return the greenlet that acquire the lock or None.
        '''
        return self._local.locked

    def acquire(self, timeout=None):
        '''Acquires the lock if in the unlocked state otherwise switch
        back to the parent coroutine.
        '''
        green = getcurrent()
        parent = green.parent
        if parent is None:
            raise RuntimeError('acquire in main greenlet')

        if self._local.locked:
            future = Future(loop=self._loop)
            self._queue.append(future)
            parent.switch(future)

        self._local.locked = green
        return self.locked()

    def release(self):
        '''Release the lock.

        This method should only be called in the locked state;
        it changes the state to unlocked and returns immediately.
        If an attempt is made to release an unlocked lock,
        a RuntimeError will be raised.
        '''
        if self._local.locked:
            while self._queue:
                future = self._queue.popleft()
                if not future.done():
                    return future.set_result(None)
            self._local.locked = None
        else:
            raise RuntimeError('release unlocked lock')

    def __enter__(self):
        self.acquire()
        return self

    def __exit__(self, type, value, traceback):
        self.release()


class GreenWSGI:
    '''Wraps a WSGI application to be executed on a :class:`.GreenPool`
    '''
    def __init__(self, wsgi, pool):
        self.wsgi = wsgi
        self.pool = pool

    def __call__(self, environ, start_response):
        return self.pool.submit(self._green_handler, environ, start_response)

    def _green_handler(self, environ, start_response):
        return wait(self.wsgi(environ, start_response))



import asyncio

from pulsar import when_monitor_start, get_application, task, send
from pulsar.apps.data import create_store
from pulsar.apps.ds import PulsarDS


def start_pulsar_ds(arbiter, host, workers=0):
    lock = getattr(arbiter, 'lock', None)
    if lock is None:
        arbiter.lock = lock = asyncio.Lock()
    yield from lock.acquire()
    try:
        app = yield from get_application('pulsards')
        if not app:
            app = PulsarDS(bind=host, workers=workers, load_config=False)
            cfg = yield from app(arbiter)
        else:
            cfg = app.cfg
        return cfg
    finally:
        lock.release()


@task
def start_store(app, url, workers=0, **kw):
    '''Equivalent to :func:`.create_store` for most cases excepts when the
    ``url`` is for a pulsar store not yet started.
    In this case, a :class:`.PulsarDS` is started.
    '''
    store = create_store(url, **kw)
    if store.name == 'pulsar':
        client = store.client()
        try:
            yield from client.ping()
        except ConnectionRefusedError:
            host = localhost(store._host)
            if not host:
                raise
            cfg = yield from send('arbiter', 'run', start_pulsar_ds,
                                  host, workers)
            store._host = cfg.addresses[0]
            dns = store._buildurl()
            store = create_store(dns, **kw)
    app.cfg.set('data_store', store.dns)


def localhost(host):
    if isinstance(host, tuple):
        if host[0] in ('127.0.0.1', ''):
            return ':'.join((str(b) for b in host))
    else:
        return host


def _start_store(monitor):
    app = monitor.app
    if not isinstance(app, PulsarDS) and app.cfg.data_store:
        start_store(app, app.cfg.data_store)

when_monitor_start.append(_start_store)



from functools import partial

from pulsar import Connection, Pool, get_actor
from pulsar.utils.pep import to_string
from pulsar.apps.data import RemoteStore
from pulsar.apps.ds import redis_parser

from .client import RedisClient, Pipeline, Consumer, ResponseError
from .pubsub import RedisPubSub


class RedisStoreConnection(Connection):

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.parser = self._producer._parser_class()

    def execute(self, *args, **options):
        consumer = self.current_consumer()
        consumer.start((args, options))
        result = yield from consumer.on_finished
        if isinstance(result, ResponseError):
            raise result.exception
        return result

    def execute_pipeline(self, commands, raise_on_error=True):
        consumer = self.current_consumer()
        consumer.start((commands, raise_on_error, []))
        result = yield from consumer.on_finished
        if isinstance(result, ResponseError):
            raise result.exception
        return result


class RedisStore(RemoteStore):
    '''Redis :class:`.Store` implementation.
    '''
    protocol_factory = partial(RedisStoreConnection, Consumer)
    supported_queries = frozenset(('filter', 'exclude'))

    def _init(self, namespace=None, parser_class=None, pool_size=50,
              decode_responses=False, **kwargs):
        self._decode_responses = decode_responses
        if not parser_class:
            actor = get_actor()
            pyparser = actor.cfg.redis_py_parser if actor else False
            parser_class = redis_parser(pyparser)
        self._parser_class = parser_class
        if namespace:
            self._urlparams['namespace'] = namespace
        self._pool = Pool(self.connect, pool_size=pool_size, loop=self._loop)
        if self._database is None:
            self._database = 0
        self._database = int(self._database)
        self.loaded_scripts = {}

    @property
    def pool(self):
        return self._pool

    @property
    def namespace(self):
        '''The prefix namespace to append to all transaction on keys
        '''
        n = self._urlparams.get('namespace')
        return '%s:' % n if n else ''

    def key(self):
        return (self._dns, self._encoding)

    def client(self):
        '''Get a :class:`.RedisClient` for the Store'''
        return RedisClient(self)

    def pipeline(self):
        '''Get a :class:`.Pipeline` for the Store'''
        return Pipeline(self)

    def pubsub(self, protocol=None):
        return RedisPubSub(self, protocol=protocol)

    def ping(self):
        return self.client().ping()

    def execute(self, *args, **options):
        connection = yield from self._pool.connect()
        with connection:
            result = yield from connection.execute(*args, **options)
            return result

    def execute_pipeline(self, commands, raise_on_error=True):
        conn = yield from self._pool.connect()
        with conn:
            result = yield from conn.execute_pipeline(commands, raise_on_error)
            return result

    def connect(self, protocol_factory=None):
        protocol_factory = protocol_factory or self.create_protocol
        if isinstance(self._host, tuple):
            host, port = self._host
            transport, connection = yield from self._loop.create_connection(
                protocol_factory, host, port)
        else:
            raise NotImplementedError('Could not connect to %s' %
                                      str(self._host))
        if self._password:
            yield from connection.execute('AUTH', self._password)
        if self._database:
            yield from connection.execute('SELECT', self._database)
        return connection

    def flush(self):
        return self.execute('flushdb')

    def close(self):
        '''Close all open connections.'''
        return self._pool.close()

    def has_query(self, query_type):
        return query_type in self.supported_queries

    def basekey(self, meta, *args):
        key = '%s%s' % (self.namespace, meta.table_name)
        postfix = ':'.join((to_string(p) for p in args if p is not None))
        return '%s:%s' % (key, postfix) if postfix else key

    def meta(self, meta):
        '''Extract model metadata for lua script stdnet/lib/lua/odm.lua'''
        #  indices = dict(((idx.attname, idx.unique) for idx in meta.indices))
        data = meta.as_dict()
        data['namespace'] = self.basekey(meta)
        return data


class CompiledQuery(object):

    def __init__(self, pipe, query):
        self.pipe = pipe



import pulsar

names = ['john', 'luca', 'jo', 'alex']


@pulsar.command()
def greetme(request, message):
    echo = 'Hello {}!'.format(message['name'])
    request.actor.logger.info(echo)
    return echo


class Greeter:

    def __init__(self):
        cfg = pulsar.Config()
        cfg.parse_command_line()
        a = pulsar.arbiter(cfg=cfg)
        self.cfg = a.cfg
        self._loop = a._loop
        self._loop.call_later(1, pulsar.async, self())
        a.start()

    def __call__(self, a=None):
        if a is None:
            a = yield from pulsar.spawn(name='greeter')
        if names:
            name = names.pop()
            self._loop.logger.info("Hi! I'm %s" % name)
            yield from pulsar.send(a, 'greetme', {'name': name})
            self._loop.call_later(1, pulsar.async, self(a))
        else:
            pulsar.arbiter().stop()


if __name__ == '__main__':
    Greeter()



from urllib.parse import urljoin

from pulsar import as_coroutine, task
from pulsar.utils.httpurl import Headers
from pulsar.utils.log import LocalMixin, local_property
from pulsar.apps.wsgi import Route, wsgi_request
from pulsar.apps.http import HttpClient


ENVIRON_HEADERS = ('content-type', 'content-length')


class Proxy(LocalMixin):
    '''Proxy requests to another server
    '''
    def __init__(self, route, url):
        self.route = Route(route)
        self.url = url

    @local_property
    def http_client(self):
        '''The :class:`.HttpClient` used by this proxy middleware for
        accessing upstream resources'''
        return HttpClient(decompress=False, store_cookies=False)

    def __call__(self, environ, start_response):
        request = wsgi_request(environ)
        path = request.path
        match = self.route.match(path[1:])
        if match is not None:
            query = request.get('QUERY_STRING', '')
            path = urljoin(self.url, match.pop('__remaining__', ''))
            if query:
                path = '%s?%s' % (path, query)
            return self._call(request, path, start_response)

    @task
    def _call(self, request, path, start_response):
        data, files = yield from as_coroutine(request.data_and_files())
        response = yield from self.http_client.request(
            request.method, path, data=data, files=files,
            headers=self.request_headers(request.environ),
            version=request.get('SERVER_PROTOCOL'))
        response.raise_for_status()
        start_response(response.get_status(), list(response.headers))
        return [response.get_content()]

    def request_headers(self, environ):
        '''Fill request headers from the environ dictionary and
        modify them via the list of :attr:`headers_middleware`.
        The returned headers will be sent to the target uri.
        '''
        headers = Headers(kind='client')
        for k in environ:
            if k.startswith('HTTP_'):
                head = k[5:].replace('_', '-')
                headers[head] = environ[k]
        for head in ENVIRON_HEADERS:
            k = head.replace('-', '_').upper()
            v = environ.get(k)
            if v:
                headers[head] = v
        return headers



'''Simple actor message passing
'''
from pulsar import arbiter, spawn, send, async, Config


def start(arbiter, **kw):
    async(app(arbiter))


def app(arbiter):
    # Spawn a new actor
    proxy = yield from spawn(name='actor1')
    print(proxy.name)
    # Execute inner method in actor1
    result = yield from send(proxy, 'run', inner_method)
    print(result)

    yield from send(proxy, 'run', set_value, 10)
    value = yield from send(proxy, 'run', get_value)
    print(value)

    # Stop the application
    arbiter.stop()


def inner_method(actor):
    return 42


def set_value(actor, value):
    actor.saved_value = value


def get_value(actor):
    return actor.saved_value


if __name__ == '__main__':
    cfg = Config()
    cfg.parse_command_line()
    arbiter(cfg=cfg, start=start).start()



'''Not used yet
'''
from pulsar import Future


class Throttling(object):
    _types = ('read', 'write')

    def __init__(self, protocol, read_limit=None, write_limit=None):
        self.protocol = protocol
        self.limits = (read_limit, write_limit)
        self.this_second = [0, 0]
        self._checks = [None, None]
        self._throttle = [None, None]

    def data_received(self, data):
        '''Receive new data and pause the receiving end of the transport
        if the raed limit per second is surpassed
        '''
        if not self._checks[0]:
            self._check_limit(0)
        self._register(data, 0)
        self.protocol.data_received(data)

    def write(self, data):
        '''Write ``data`` and return either and empty tuple or
        an :class:`~asyncio.Future` called back once the write limit
        per second is not surpassed.
        '''
        if not self._checks[1]:
            self._check_limit(1)
        self._register(data, 1)
        result = self.protocol.write(data)
        waiter = self._throttle[1]
        return self.gather(result, waiter)

    # INTERNALS
    def _register(self, data, rw):
        self.this_second[rw] += len(data)

    def _check_limit(self, rw):
        loop = self.protocol._loop
        limit = self.limits[rw]
        if limit and self.this_second[rw] > limit:
            self._thorttle(rw)
            next = (float(self.this_second[rw]) / limit) - 1.0
            loop.call_later(next, self._un_throttle, rw)
        self.this_second[rw] = 0
        self._checks[rw] = loop.call_later(1, self._check_limit, rw)

    def _throttle(self, rw):
        self.logger.debug('Throttling %s', self._types[rw])
        if rw:
            assert not self._throttle[rw]
            self._throttle[rw] = Future(self.protocol._loop)
        else:
            self._throttle[rw] = True
            t = self.protocol._transport
            t.pause_reading()

    def _un_throttle(self, rw):
        self.protocol.logger.debug('Un-throttling %s', self._types[rw])
        if rw:
            waiter = self._throttle[rw]
            self._throttle[rw] = None
            if waiter and not waiter.done():
                waiter.set_result(None)
        else:
            t = self.protocol._transport
            if t:
                t.resume_reading()



from django.conf.urls import patterns, include, url
from django.contrib import admin
from django.contrib.staticfiles.urls import staticfiles_urlpatterns


urlpatterns = patterns(
    '',
    url(r'^$', 'djchat.views.home'),
    url(r'^admin/', include(admin.site.urls)),
)

urlpatterns += staticfiles_urlpatterns()
